# Copyright 2023 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import hashlib
import os
import webbrowser
from multiprocessing import Queue

import requests
import uvicorn
from fastapi import APIRouter, FastAPI, Request
from fastapi.responses import RedirectResponse
from ipyauth import Auth, ParamsAuth0

from oqtant.settings import Settings
from oqtant.util.server import ThreadServer

settings = Settings()

app = FastAPI(title="Login API", openapi_url="/openapi.json")
router = APIRouter()


def generate_random(length: int) -> str:
    """Method to generate a random base64 string

    Args:
        length (int): The length to make the string

    Returns:
        str: The generated base64 string
    """
    return base64.urlsafe_b64encode(os.urandom(length)).decode("utf-8").replace("=", "")


verifier = generate_random(80)


def generate_challenge(verifier: str) -> str:
    """Method to generate a base64 string to serve as an auth0 challenge

    Args:
        verifier (str): A base64 string to create the challenge off of

    Returns:
        str: The generated challenge
    """
    hashed = hashlib.sha256(verifier.encode("utf-8")).digest()
    return base64.urlsafe_b64encode(hashed).decode("utf-8").replace("=", "")


def get_authentication_url(auth_server_port: int) -> None:
    """Method to generate the auth0 authentication url

    Args:
        auth_server_port (int): The port of the server auth0 should redirect to
    """
    code_challenge = generate_challenge(verifier)
    auth_url = "".join(
        [
            f"https://{settings.auth0_domain}/authorize",
            "?response_type=code",
            f"&scope={settings.auth0_scope}",
            f"&audience={settings.auth0_audience}",
            f"&code_challenge={code_challenge}",
            "&code_challenge_method=S256",
            f"&client_id={settings.auth0_client_id}",
            f"&redirect_uri=http://localhost:{auth_server_port}",
        ]
    )
    return auth_url


queue = Queue()


@app.get("/")
async def main(request: Request, code: str) -> str:
    """Main route to handle user authentication

    Args:
        request (fastapi.Request): The request object
        code (str): The code provided by auth0 to be verified

    Returns:
        str: Message on whether the user successfully authenticated
    """
    resp = await get_token(verifier, code, request.url.port)
    token = resp["access_token"]
    queue.put({"token": token})
    if token:
        return "Successfully authenticated, you may close this tab now"
    else:
        return "Failed to authenticate, please close this tab and try again"


@app.get("/login")
def login(request: Request) -> RedirectResponse:
    """Route to initiate the authentication process

    Args:
        request (Request): The request object

    Returns:
        RedirectResponse: A redirect to auth0 universal login
    """
    return RedirectResponse(
        url=get_authentication_url(auth_server_port=request.url.port)
    )


async def get_token(verifier: str, code: str, auth_server_port: int) -> dict:
    """Method to get an authentication token from auth0 after a user authenticates

    Args:
        verifier (str): The base64 string to provide along with the code challenge
        code (str): The code challenge returned from auth0
        auth_server_port (int): The port the local server is running on

    Returns:
        dict: The json response from auth0, should contain the authentication token
    """
    url = f"https://{settings.auth0_domain}/oauth/token"
    headers = {"content-type": "application/x-www-form-urlencoded"}
    data = {
        "grant_type": "authorization_code",
        "client_id": settings.auth0_client_id,
        "code_verifier": verifier,
        "code": code,
        "redirect_uri": f"http://localhost:{auth_server_port}",
    }
    resp = requests.post(
        url, headers=headers, data=data, allow_redirects=False, timeout=(5, 30)
    )
    return resp.json()


def get_user_token(auth_server_port: int = 8080) -> str:
    """Method to initiate the user authentication process

    Args:
        auth_server_port (int, optional): The port to run the server on locally

    Returns:
        str: The authentication token
    """
    allowed_ports = [8080, 8081, 8082, 8083, 8084, 8085]
    if auth_server_port not in allowed_ports:
        raise ValueError(f"{auth_server_port} not in allowed ports: {allowed_ports}")
    server_config = uvicorn.Config(
        app=app, host="localhost", port=auth_server_port, log_level="error"
    )
    server = ThreadServer(config=server_config)
    with server.run_in_thread():
        webbrowser.open(f"http://localhost:{auth_server_port}/login")
        token = queue.get(block=True)
    return token["token"]


def notebook_login() -> Auth:
    """Method to get an authenticate widget

    Returns:
        ipyauth.Auth: The authentication widget
    """
    p = ParamsAuth0(
        domain=settings.auth0_domain,
        client_id=settings.auth0_client_id,
        redirect_uri=settings.auth0_notebook_redirect_uri,
        audience=settings.auth0_audience,
        response_type=settings.auth0_response_type,
        scope=settings.auth0_scope,
    )
    a = Auth(params=p)
    return a



# Copyright 2023 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bert_schemas import job as job_schema
from scipy.interpolate import interp1d


def interpolation_to_kind(interpolation: job_schema.InterpolationType) -> str:
    """Method to convert our InterpolationType to something scipy can understand

    Args:
        interpolation (bert_schemas.job.InterpolationType): Primitive job interpolation type

    Returns:
        str: A "kind" string to be used by scipy's interp1d
    """
    interpolation_map = {"OFF": "zero", "STEP": "previous", "SMOOTH": "cubic"}

    return interpolation_map.get(interpolation, interpolation.lower())


def interpolate_1d(
    xs: list[float],
    ys: list[float],
    x: float,
    interpolation: job_schema.InterpolationType = "LINEAR",
) -> float:
    """Method to interpolate a 1D list of pairs [xs, ys] at the evaluation point x

    Args:
        xs (list[float]): List of x values
        ys (list[float]): List of y values
        x (float): Desired x-coordinate to evaluate the resulting interpolation function
        interpolation (job_schema.InterpolationType, optional): Interpolation style

    Returns:
        float: Interpolation function value at the specified x-coordinate
    """
    f = interp1d(
        xs,
        ys,
        kind=interpolation_to_kind(interpolation),
        bounds_error=False,
        fill_value=(0.0, 0.0),
    )
    return f(x)[()]  # extract value


def interpolate_1d_list(
    xs: list[float],
    ys: list[float],
    x_values: list[float],
    interpolation: job_schema.InterpolationType = "LINEAR",
) -> list[float]:
    """Method to interpolate a 1d list of pairs [xs, ys] at the evaluation points given by x_values

    Args:
        xs (list[float]): List of x values
        ys (list[float]): List of y values
        x_values (list[float]): Desired x-coordinates to evaluate the resulting interpolation function
        interpolation (job_schema.InterpolationType, optional): Interpolation style

    Returns:
        list[float]: Floating point values corresponding to evaluation of the interpolation function
            value at the specified x_values
    """
    f = interp1d(
        xs,
        ys,
        kind=interpolation_to_kind(interpolation),
        bounds_error=False,
        fill_value=(0.0, 0.0),
    )
    return list(f(x_values))



# Copyright 2023 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from dataclasses import dataclass
from logging import getLogger

import matplotlib.pyplot as plt
import numpy as np
from bert_schemas.job import Image
from scipy import interpolate

from oqtant.simulator.qm_potential import QMPotential
from oqtant.simulator.three_dim import ThreeDim
from oqtant.simulator.wave_function import WaveFunction
from oqtant.util.exceptions import SimValueError

logger = getLogger(__name__)


@dataclass
class TimeSpan:
    start: float
    end: float


class Simulator:
    """
    'Simulator' Defines methods for evolution and plotting of the system
    """

    # 2000 timesteps per simulation unit time.
    # time_step = 1 #0.5e-3
    number_of_atoms = 1e5
    # this is specific to interactions between Rubidium - 87 atoms
    scattering_length = 0.00319
    interaction_strength = 4.0 * np.pi * number_of_atoms * scattering_length

    def __init__(self, wavefunction: WaveFunction, potential: QMPotential):
        self.tdm = ThreeDim()
        self.wavefunction = wavefunction
        self.potential = potential
        self.times: list | None = None
        self.psi_history = []  # None
        self.time_step = 0.5e-3
        self.sampling_rate = 100

    def set_ground_state(self):
        """
        This function evolves the condensate with ground = True.
        It is done at negative times before the barriers are switched on.
        2.5 simulation units of time is sufficient to settle down to the ground state.
        Potentially this could be stored as it's always the same unless the user wants to change number_of_atoms.
        """

        self.wavefunction.psi = self.wavefunction.initial_psi(
            sigma_r=1 / np.sqrt(self.potential.w0r)
        )

        self.run_RK4(TimeSpan(-2.5, -0.01), ground=True)

    def run_evolution(self) -> None:
        """
        This function evolves the condensate with ground = False.
        It is done at positive times while the barriers are switched on.
        It runs for the lifetime of the quantum_matter object
        It starts from the end result of running get_ground_state
        """
        # start from the end result of running get_ground_state
        self.wavefunction.psi = self.psi_history[-1]
        self.run_RK4(TimeSpan(0, self.potential.lifetime), ground=False)

    # Move this to a separate class of utility functions (separate module)

    def get_laplacian(self, y) -> np.ndarray:
        """
        Implementation of the second derivatives in x and r including forward, central, and backward formulas to second order accuracy
        """
        # dr, dx = self.wavefunction.dr, self.wavefunction.dx
        dr, dx = self.tdm.dr, self.tdm.dx

        # First derivative.
        dydr = np.gradient(y, axis=0, edge_order=2) / dr  # First derivative.
        dydr[0, :] = 0  # Enforce Neumann boundary condition

        #         # Central difference
        d2ydr2 = (np.roll(y, -1, axis=0) - 2 * y + np.roll(y, 1, axis=0)) / dr**2
        d2ydx2 = (np.roll(y, -1, axis=1) - 2 * y + np.roll(y, 1, axis=1)) / dx**2

        #         # Forward difference at start of array
        d2ydr2[0, :] = (2 * y[0, :] - 5 * y[1, :] + 4 * y[2, :] - y[3, :]) / dr**2
        d2ydx2[:, 0] = (2 * y[:, 0] - 5 * y[:, 1] + 4 * y[:, 2] - y[:, 3]) / dx**2

        #         # Backward difference at end of array
        d2ydr2[-1, :] = (
            -y[-4, :] + 4 * y[-3, :] - 5 * y[-2, :] + 2 * y[-1, :]
        ) / dr**2
        d2ydx2[:, -1] = (
            -y[:, -4] + 4 * y[:, -3] - 5 * y[:, -2] + 2 * y[:, -1]
        ) / dx**2

        return (
            (1.0 / self.tdm.r) * dydr + d2ydr2 + d2ydx2
        )  # Laplacian in cylindrical coordinates with axial symmetry

    def get_GPE(self, psi: np.ndarray) -> np.ndarray:
        """
        Implementation of the Gross-Pitaevskii Equation w/Neumann boundary conditions at r = 0 and Dirichlet at large x and r.
        """
        # Enforce Neumann boundary condition
        psi[1, :] = psi[0, :]

        laplacian = self.get_laplacian(psi)

        return (0.5 * 1j) * laplacian - 1j * (
            self.potential.V + self.interaction_strength * np.abs(psi) ** 2
        ) * psi

    def run_RK4(self, time_span: TimeSpan, ground=False, TOF=False) -> None:
        """
        Implementation of the Runge-Kutta 4th order method to evolve in time.

        Args:
            time_span:  a list of times (in milliseconds)
            ground: allows simulation of ground state and evolves in imaginary time
            TOF:  whether or not to switches off potentials and rescale the grids
        Returns:
            None
        """
        n = int(
            (time_span.end - time_span.start) / self.time_step
        )  # number of subintervals to evaluate rk4 on.
        t = np.linspace(
            time_span.start, time_span.end, n + 1
        )  # list of all time subintervals
        self.times = []  # stored internally in an array
        self.psi_history = []  # reset history each time rk4 is called

        f = np.zeros(
            (self.tdm.Nr, self.tdm.Nx),
            dtype=complex,
        )  # stored wave function at each time

        # Set the initial wave function
        f = self.wavefunction.psi

        dt = self.time_step

        # if ground = True, we evolve in imaginary time
        if ground:
            dt = -dt * 1j

        oqt_times_us = t * QMPotential.sim_time_to_oqt / QMPotential.msec

        for i in range(n):
            if i == 0:
                oqt_times_us[i]
                self.potential.update_potential(t[i])

            elif oqt_times_us[i] % 100 < oqt_times_us[i - 1] % 100 and ground is False:
                self.potential.update_potential(t[i])

            psi = self.wavefunction.psi

            f1 = self.get_GPE(psi)
            f2 = self.get_GPE(psi + f1 * dt / 2)
            f3 = self.get_GPE(psi + f2 * dt / 2)
            f4 = self.get_GPE(psi + f3 * dt)

            f = f + dt * (f1 + 2.0 * f2 + 2.0 * f3 + f4) / 6.0

            # Dirichlet boundary conditions at asymptotes of simulation
            f[-1, :] = 0
            f[:, 0] = 0
            f[:, -1] = 0

            # Enforce Neumann boundary condition
            f[1, :] = f[0, :]

            self.wavefunction.psi = f

            if TOF:  # interpolate wave function between expanding grids
                self.TOF_mode(i, interp_psi=True)
                f = self.wavefunction.psi

            if ground:  # renormalize wave function due to imaginary time decay
                self.wavefunction.psi = self.wavefunction.normalize(
                    self.wavefunction.psi
                )
                f = self.wavefunction.psi

            if i % self.sampling_rate == 0:
                logger.info("sample", t[i])
                self.psi_history.append(f)
                self.times.append(t[i])

        logger.info("psi_history", np.shape(self.psi_history))

    @property
    def it_plot(self) -> Image:
        """
        Generate an simulation analog to an in-trap image from the Oqtant hardware.

        Args:

        Returns:
        - it_image_pixels (list): pixels of the simulation generated IT image
        - columns (int): number of cols in simulation generated IT images
        - rows (int): number of rows in simulation generated IT images
        - pixcal (float): size of a pixel in simulation generated IT images
        """

        # get the Z-integrated column denisty from wavefunction
        _, density_img, _ = self.wavefunction.column_densities
        density_pixels = density_img.flatten() * self.tdm.dx * self.tdm.dz

        # get conversion from simulation length to um
        pixcal = QMPotential.sim_length_to_oqt

        A_pixel_area = (pixcal) ** 2  # um2

        # absorption cross section
        sigma = 2.907e-1  # um2

        it_exp_pixcal = 0.344  # um

        # convert density (atoms/length^2) to OD using the sim grid size and cross section
        OD_pixels = density_pixels * sigma * self.number_of_atoms / A_pixel_area

        # half the length of the simulation grid axes, in um
        Lx_half = (self.tdm.Lx / 2) * pixcal
        Ly_half = self.tdm.Lr * pixcal

        # grids for building simulation interpolation function (um)
        sim_grid_col = np.linspace(-Lx_half, Lx_half, self.tdm.Nx)
        sim_grid_row = np.linspace(-Ly_half, Ly_half, self.tdm.Nr * 2)

        # build interpolation function for sim data
        f = interpolate.RegularGridInterpolator(
            (sim_grid_row, sim_grid_col),
            OD_pixels.reshape((self.tdm.Nr * 2, self.tdm.Nx)),
        )

        # grids for sampling simulation in experiment resolution (um)
        exp_grid_col = np.arange(-Lx_half, Lx_half, it_exp_pixcal)
        exp_grid_row = np.arange(-Ly_half, Ly_half, it_exp_pixcal)

        exp_meshgrid_row, exp_meshgrid_col = np.meshgrid(exp_grid_row, exp_grid_col)

        points = np.array([exp_meshgrid_row.flatten(), exp_meshgrid_col.flatten()]).T

        # sample the interpolation function and reshape the output
        interpolated_sim = f(points).reshape(len(exp_grid_col), len(exp_grid_row))

        # how many pixels to pad the image?
        add_zeros_rows = math.ceil((148 - len(exp_grid_row)) / 2)
        add_zeros_cols = math.ceil((512 - len(exp_grid_col)) / 2)

        # pad the image to match IT_PLOT size
        padded_interp_sim = np.pad(
            interpolated_sim,
            [
                [add_zeros_cols, 512 - len(exp_grid_col) - add_zeros_cols],
                [add_zeros_rows, 148 - len(exp_grid_row) - add_zeros_rows],
            ],
        ).T

        # prepare values to be returned
        columns = len(padded_interp_sim[0])
        rows = len(padded_interp_sim)
        pixels_pad_interp_sim = list(padded_interp_sim.flatten())

        return Image(
            pixels=pixels_pad_interp_sim,
            rows=rows,
            columns=columns,
            pixcal=it_exp_pixcal,
        )

    def show_final_result(self) -> None:
        """
        Plot the density at the end of the simulation in cylindrical coordinates.
        Useful coordinates for diagonising issues but not to be returned to the user.
        """
        fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
        density = self.wavefunction.density
        ax.plot_surface(
            self.potential.sim_length_to_oqt * self.tdm.r,
            self.potential.sim_length_to_oqt * self.tdm.x,
            density,
            cmap="viridis",
        )
        ax.set_xlabel("$R$-Position (microns)")
        ax.set_ylabel("$X$-Position (microns)")

    # @staticmethod
    def convert_timesteps(self, timesteps: list) -> np.ndarray:
        if self.times is None:
            raise SimValueError("times not set")

        np_timesteps = np.array(timesteps, dtype=float)
        # convert input time in msec to simulation units
        np_timesteps *= QMPotential.oqt_time_to_sim

        if (
            np.max(np_timesteps) >= self.times[-1]
            or np.min(np_timesteps) <= self.times[0]
        ):
            raise SimValueError("timesteps outside of job window")

        if np.min(np_timesteps) < 0:
            raise SimValueError("timesteps must be positive")

        return np_timesteps

    def show_density_cylindrical(self, timesteps: list, TOF=False) -> None:
        """
        Plots the density profile of the condensate in cylindrical coordinates
        for an input array of times.
        Useful coordinates for diagonising issues but not to be returned to the user.

        Args:
            timesteps:  a list of times (in milliseconds)
            TOF:  whether or not to switch off potentials and rescale the grids
        Returns:
            None
        """

        converted_timesteps = self.convert_timesteps(timesteps)

        # convert to simulation indices for the nearest timestep, loop over timesteps
        for time in converted_timesteps:
            i = int(time / (self.time_step * self.sampling_rate))

            # if TOF:  # rescale grids in TOF mode.
            #     self.TOF_mode(i)

            self.potential.update_potential(self.times[i])

            fig = plt.figure(figsize=(15, 7))

            ax1 = fig.add_subplot(
                111
            )  # Add first subplot in a plot with two rows and 2 columns.
            self.wavefunction.psi = self.psi_history[i]
            density = self.wavefunction.density
            img1 = ax1.imshow(
                np.concatenate(
                    (np.flip(density, axis=0), density)
                ),  # mirror the condensate over the z-axis
                extent=[
                    self.tdm.x[0, 0] * self.potential.sim_length_to_oqt,
                    self.tdm.x[0, -1] * self.potential.sim_length_to_oqt,
                    -self.tdm.r[-1, 0] * self.potential.sim_length_to_oqt,
                    self.tdm.r[-1, 0] * self.potential.sim_length_to_oqt,
                ],
                aspect="auto",
                cmap="viridis",
            )
            plt.title(f"Density at time {time / self.potential.oqt_time_to_sim} msec")
            ax1.set_xlabel("$X$-Position (microns)")
            ax1.set_ylabel("$R$-Position (microns)")
            ax1.contour(
                self.tdm.x,
                self.tdm.r,
                self.potential.V,
                colors="white",
            )
            ax1.contour(
                self.tdm.x,
                -self.tdm.r,
                self.potential.V,
                colors="white",
            )
            fig.colorbar(img1, ax=ax1, orientation="vertical")

    def show_column_densities(self, timesteps, TOF=False):
        """
        Plots the column densities and slices of the condensate in cartesian coordinates
        for an input array of times.
        In correct coordinates to be returned to the user.

        Args:
            timesteps:  a list of times (in milliseconds)
            TOF:  whether or not to rescale the grids in TOF mode

        Returns:
            None
        """
        converted_timesteps = self.convert_timesteps(timesteps)

        for time in converted_timesteps:
            i = int(time / (self.time_step * self.sampling_rate))

            # if TOF:  # rescale grids in TOF mode.
            #     self.TOF_mode(i)

            self.potential.update_potential(
                self.times[i]
            )  # update potential to plotted time

            self.wavefunction.psi = self.psi_history[i]  # load a snapshot of psi
            # load the column densities
            (
                column_zy,
                column_zx,
                profiles,
            ) = self.wavefunction.column_densities
            slice_x_2 = column_zx[int(self.tdm.Nr - 1), :]
            slice_z = column_zy[int(self.tdm.Nr - 1), :]
            slice_z_zy, slice_z_zx, slice_x = profiles

            fig = plt.figure(figsize=(15, 7))

            # column densities
            ax1 = fig.add_subplot(
                221
            )  # Add first subplot in a plot with two rows and 2 columns.
            im1 = ax1.imshow(
                column_zy,
                extent=[
                    -self.tdm.Lr * self.potential.sim_length_to_oqt,
                    self.tdm.Lr * self.potential.sim_length_to_oqt,
                    -self.tdm.Lr * self.potential.sim_length_to_oqt,
                    self.tdm.Lr * self.potential.sim_length_to_oqt,
                ],
                aspect="auto",
                cmap="viridis",
            )
            plt.title(
                f"Column densities at time {time / self.potential.oqt_time_to_sim} msec"
            )
            plt.xlabel("$Y$-Position (microns)")
            plt.ylabel("$Z$-Position (microns)")
            fig.colorbar(im1, ax=ax1, orientation="vertical")

            ax2 = fig.add_subplot(222)
            im2 = ax2.imshow(
                column_zx,
                extent=[
                    -self.potential.sim_length_to_oqt * self.tdm.Lx / 2,
                    self.potential.sim_length_to_oqt * self.tdm.Lx / 2,
                    -self.potential.sim_length_to_oqt * self.tdm.Lr,
                    self.potential.sim_length_to_oqt * self.tdm.Lr,
                ],
                aspect="auto",
                cmap="viridis",
            )
            plt.xlabel("$X$-Position (microns)")
            plt.ylabel("$Z$-Position (microns)")
            fig.colorbar(im2, ax=ax2, orientation="vertical")

            # slices
            ax3 = fig.add_subplot(223)
            ax3.plot(self.tdm.z[0, :] * self.potential.sim_length_to_oqt, slice_z)
            ax3.set_title(
                f"Slices at time {time / self.potential.oqt_time_to_sim} msec"
            )
            ax3.set_xlabel("$Z$-Position (microns)")

            #           #######TBD ADD show_potential output here!
            ax4 = fig.add_subplot(224)
            ax4.plot(self.tdm.x[0, :] * self.potential.sim_length_to_oqt, slice_x_2)
            ax4.set_xlabel("$X$-Position (microns)")

            plt.tight_layout()
            plt.show()

    def show_phase(self, timesteps, TOF=False) -> None:
        """
        Plot the phase for a given list of timesteps
        This can only be displayed in cylindrical coordinates.
        It is a helpful tool still for the user.  The aspect ratio is still a bit weird.

        Args:
            timesteps (list): List of times to display
            TOF (bool): True to rescale grids in TOF mode.
        Returns:
            None
        """
        converted_timesteps = self.convert_timesteps(timesteps)

        for time in converted_timesteps:
            # convert to simulation indices for the nearest timestep, loop over timesteps
            i = int(time / (self.time_step * self.sampling_rate))

            # if TOF:  # rescale grids in TOF mode.
            #     self.TOF_mode(i)

            self.potential.update_potential(self.times[i])

            self.wavefunction.psi = self.psi_history[i]

            plt.figure(figsize=(10, 7))
            plt.imshow(
                self.wavefunction.phase,
                extent=[
                    self.tdm.x[0, 0] * self.potential.sim_length_to_oqt,
                    self.tdm.x[0, -1] * self.potential.sim_length_to_oqt,
                    0,
                    self.tdm.r[-1, 0] * self.potential.sim_length_to_oqt,
                ],
                aspect="auto",
            )
            plt.colorbar(label="Phase")
            plt.clim(-np.pi, np.pi)
            plt.title(f"Phase at time {time / self.potential.oqt_time_to_sim} msec")
            plt.xlabel("$X$-Position (microns)")
            plt.ylabel("$R$-Position (microns)")
            plt.show()

    def show_current(self, timesteps, TOF=False):
        """
        Plot the flow for a given list of timesteps
        two separate sub plots

        Args:
            timesteps (list): List of times to display
            TOF (bool): True to rescale grids in TOF mode.

        Returns:
            None
        """
        converted_timesteps = self.convert_timesteps(timesteps)

        for time in converted_timesteps:
            i = int(time / (self.time_step * self.sampling_rate))

            # if TOF:  # rescale grids in TOF mode.
            #     self.TOF_mode(i)

            self.potential.update_potential(self.times[i])

            self.wavefunction.psi = self.psi_history[i]

            plt.figure(1)
            plt.plot(
                self.tdm.x_1d * self.potential.sim_length_to_oqt,
                self.number_of_atoms
                * self.wavefunction.current
                / self.potential.sim_time_to_oqt,
            )
            plt.title(f"Current at time {time / self.potential.oqt_time_to_sim} msec")
            plt.xlabel("$X$-Position (microns)")
            plt.ylabel("Atom Current (# of Atoms / msec)")
            plt.show()


#    # TBD Include flow_r in plot also.
#    def animate_flow(self, frame_interval=50):
#        """
#        Animate the change in flow
#        int frame_interval: number of frames to skip each interval, determines smoothness
#        """
#        fig, ax = plt.subplots()
#        flow_r, flow_z = self.wavefunction.flow
#        density = self.wavefunction.density
#        extent = extent = [
#            self.wavefunction.z[0, 0],
#            self.wavefunction.z[0, -1],
#            -self.wavefunction.r[-1, 0],
#            self.wavefunction.r[-1, 0],
#        ]
#        im = ax.imshow(
#            np.concatenate((np.flip(flow_z * density, axis=0), flow_z * density)),
#            aspect="auto",
#            extent=extent,
#            cmap="seismic",
#        )

#        cont1 = ax.contour(
#            self.wavefunction.z,
#            self.wavefunction.r,
#            self.potential.harmonicPotential(),
#            colors="white",
#        )
#        cont2 = ax.contour(
#            self.wavefunction.z,
#            -self.wavefunction.r,
#            self.potential.harmonicPotential(),
#            colors="white",
#        )
#        levels = [self.potential.barrier_height / 2]
#        cont3 = ax.contour(
#            self.wavefunction.z,
#            self.wavefunction.r,
#            self.potential.painted_barrier(),
#            levels,
#            colors="red",
#        )
#        cont4 = ax.contour(
#            self.wavefunction.z,
#            -self.wavefunction.r,
#            self.potential.painted_barrier(),
#            levels,
#            colors="red",
#        )
#        plt.colorbar(im, ax=ax, label="current - z")
#        im.set_clim(-1, 1)

#        def update(frame):
#            nonlocal cont1, cont2, cont3, cont4
#            self.potential.adjust_potential(self.times[frame])
#            self.wavefunction.psi = self.psi_history[frame]
#            density = self.wavefunction.density
#            flow_r, flow_z = self.wavefunction.flow
#            im.set_array(
#                np.concatenate((np.flip(flow_z * density, axis=0), flow_z * density))
#            )
#            for c in cont1.collections:
#                c.remove()
#            for c in cont2.collections:
#                c.remove()
#            for c in cont3.collections:
#                c.remove()
#            for c in cont4.collections:
#                c.remove()
#            cont1 = ax.contour(
#                self.wavefunction.z,
#                self.wavefunction.r,
#                self.potential.harmonicPotential(),
#                colors="white",
#            )
#            cont2 = ax.contour(
#                self.wavefunction.z,
#                -self.wavefunction.r,
#                self.potential.harmonicPotential(),
#                colors="white",
#            )
#            levels = [self.potential.barrier_height / 2]
#            cont3 = ax.contour(
#                self.wavefunction.z,
#                self.wavefunction.r,
#                self.potential.painted_barrier(),
#                levels,
#                colors="red",
#            )
#            cont4 = ax.contour(
#                self.wavefunction.z,
#                -self.wavefunction.r,
#                self.potential.painted_barrier(),
#                levels,
#                colors="red",
#            )
#            #
#            return im, cont1, cont2, cont3, cont4

#        anim = FuncAnimation(
#            fig,
#            update,
#            frames=range(0, len(self.psi_history), frame_interval),
#            interval=50,
#            blit=False,
#        )

#        return anim

# This function is specific to doing time-of-flight.  We will implement after getting in-trap to work.  The coordinates need to be updated.
#    def TOF_mode(self, time, interp_psi=False):
#        """
#        Switches off the confining potential and the barriers.
#        Scales the grids and grid spacings in time as the gas is released from the trap.
#        Interpolates the wave function onto the rescaled grids at each time step.
#        time:  expansion time
#        interp_psi:  if 'True', the wave function is interpolated between grids.
#        set to 'False' to dilate just grids for plotting purposes.
#        """
#        # timestamp
#        t = self.times[time]

#        # confining potentials and barriers are switched off.
#        self.potential.w0z, self.potential.w0r, self.potential.barrier_height = [
#            0,
#            0,
#            0,
#        ]

#        # a fixed dilation factor.   #####CHANGE THIS FOR CIGAR IN LONG DIRECTION!
#        if self.interaction_strength == 0:
#            vz, vr = [
#                2 * np.sqrt(2 * self.w0z),
#                np.sqrt(2 * self.w0r),
#            ]  # ideal condensate expansion rates
#        else:
#            vz, vr = [
#                2 * self.R_tf_x * self.w0z,
#                self.R_tf_y * self.w0r,
#            ]  # TF expansion rates
#            # vx, vy = [8, 2 * self.R_tf_y * self.w0y] # ideal in x, TF in y

#        # rescale grid lengths linearly
#        self.potential.Lz = self.Lz + t * vz  # grid lengths expand at rates vx, vy
#        self.potential.Lr = self.Lr + t * vr

#        # dilate grids
#        self.potential.z = self.z * (self.potential.Lz / self.Lz)
#        self.potential.r = self.r * (self.potential.Lr / self.Lr)

#        # rescale grid-spacings
#        self.potential.dz = self.potential.z[0, 1] - self.potential.z[0, 0]
#        self.potential.dr = self.potential.r[1, 0] - self.potential.r[0, 0]

#        # interpolating function for the real and imaginary parts of the wave function.  Keep at least cubic order in the interpolator.
#        if interp_psi:
#            ip_real = RectBivariateSpline(
#                self.wavefunction.r[:, 0],
#                self.wavefunction.z[0, :],
#                np.real(self.wavefunction.psi),
#                kx=3,
#                ky=3,
#            )  # cubic spline interpolation
#            ip_imag = RectBivariateSpline(
#                self.wavefunction.r[:, 0],
#                self.wavefunction.z[0, :],
#                np.imag(self.wavefunction.psi),
#                kx=3,
#                ky=3,
#            )  # cubic spline interpolation

#        # dilate wave function class grids
#        self.wavefunction.r, self.wavefunction.z = self.potential.r, self.potential.z
#        self.wavefunction.Lr, self.wavefunction.Lz = (
#            self.potential.Lr,
#            self.potential.Lz,
#        )
#        self.wavefunction.dr, self.wavefunction.dz = (
#            self.potential.dr,
#            self.potential.dz,
#        )

#        # wave function interpolated onto dilated grids
#        if interp_psi:
#            self.wavefunction.psi = ip_real(
#                self.wavefunction.r[:, 0], self.wavefunction.z[0, :]
#            ) + 1j * ip_imag(self.wavefunction.r[:, 0], self.wavefunction.z[0, :])

#    def animate_phase(self, frame_interval=15):
#        """
#        Animate the change in phase
#        int frame_interval: number of frames to skip each interval, determines smoothness
#        """
#        fig, ax = plt.subplots()
#        phase = self.wavefunction.phase

#        extent = [
#            self.wavefunction.z[0, 0],
#            self.wavefunction.z[0, -1],
#            -self.wavefunction.r[-1, 0],
#            self.wavefunction.r[-1, 0],
#        ]
#        im = ax.imshow(
#            np.concatenate((np.flip(phase, axis=0), phase)),
#            extent=extent,
#            vmin=-np.pi,
#            vmax=np.pi,
#            cmap="nipy_spectral",
#        )

#        cont1 = ax.contour(
#            self.wavefunction.z,
#            self.wavefunction.r,
#            self.potential.harmonicPotential(),
#            colors="white",
#        )
#        cont2 = ax.contour(
#            self.wavefunction.z,
#            -self.wavefunction.r,
#            self.potential.harmonicPotential(),
#            colors="white",
#        )
#        levels = [self.potential.barrier_height / 2]
#        cont3 = ax.contour(
#            self.wavefunction.z,
#            self.wavefunction.r,
#            self.potential.painted_barrier(),
#            levels,
#            colors="red",
#        )
#        cont4 = ax.contour(
#            self.wavefunction.z,
#            -self.wavefunction.r,
#            self.potential.painted_barrier(),
#            levels,
#            colors="red",
#        )
#        plt.colorbar(im, ax=ax, label="phase")

#        def update(frame):
#            nonlocal cont1, cont2, cont3, cont4
#            self.potential.adjust_potential(self.times[frame])
#            self.wavefunction.psi = self.psi_history[frame]
#            phase = self.wavefunction.phase()
#            im.set_array(np.concatenate((np.flip(phase, axis=0), phase)))
#            for c in cont1.collections:
#                c.remove()
#            for c in cont2.collections:
#                c.remove()
#            for c in cont3.collections:
#                c.remove()
#            for c in cont4.collections:
#                c.remove()
#            cont1 = ax.contour(
#                self.wavefunction.z,
#                self.wavefunction.r,
#                self.potential.harmonicPotential(),
#                colors="white",
#            )
#            cont2 = ax.contour(
#                self.wavefunction.z,
#                -self.wavefunction.r,
#                self.potential.harmonicPotential(),
#                colors="white",
#            )
#            levels = [self.potential.barrier_height / 2]
#            cont3 = ax.contour(
#                self.wavefunction.z,
#                self.wavefunction.r,
#                self.potential.painted_barrier(),
#                levels,
#                colors="red",
#            )
#            cont4 = ax.contour(
#                self.wavefunction.z,
#                -self.wavefunction.r,
#                self.potential.painted_barrier(),
#                levels,
#                colors="red",
#            )
#            #
#            return im, cont1, cont2, cont3, cont4

#        anim = FuncAnimation(
#            fig,
#            update,
#            frames=range(0, len(self.psi_history), frame_interval),
#            interval=50,
#            blit=False,
#        )

#        return anim

# A useful function for when we do time-of-flight, but not needed now.
#
#    # Evolution of the widths and COM of the condensate
#    def plot_widths(self, TOF=False, COM=False):
#        """
#        Plots the change in the condensate width over time.  Comparison is made against the width of
#        a non-interacting BEC.  If TOF = True, we compare against the expansion of a non-interacting and TF BEC.
#        """

#        width_z = np.zeros(len(self.psi_history))
#        width_r = np.zeros_like(width_z)
#        width_z_ideal = np.zeros_like(width_z)
#        width_r_ideal = np.zeros_like(width_z)
#        R_z_TF = np.zeros_like(width_z)
#        R_r_TF = np.zeros_like(width_z)
#        com_z = np.zeros_like(width_z)

#        for i in range(0, len(self.psi_history)):
#            self.wavefunction.psi = self.psi_history[i]
#            if TOF:  # rescale grids in TOF mode.
#                self.TOF_mode(i)
#                scale_r = np.sqrt(1 + self.w0r**2 * self.times[i] ** 2)
#                scale_z = np.sqrt(1 + self.w0z**2 * self.times[i] ** 2)
#                width_r_ideal[i] = (
#                    self.width_r_ideal * scale_r
#                )  # non-interacting widths
#                width_z_ideal[i] = self.width_z_ideal * scale_z
#                R_r_TF[i] = self.R_tf_r * scale_r
#                R_z_TF[i] = self.R_tf_z * scale_z

#            else:
#                width_z_ideal[i], width_r_ideal[i] = (
#                    self.width_z_ideal,
#                    self.width_r_ideal,
#                )
#                R_z_TF[i], R_r_TF[i] = self.R_tf_z, self.R_tf_r

#            width_z[i], width_r[i] = self.wavefunction.widths()
#            com_z[i] = self.wavefunction.com_position()

#        TF_z_to_width = np.sqrt(
#            1 / 7
#        )  # Conversion between thomas fermi radii and widths
#        TF_r_to_width = np.sqrt(2 / 7)

#        # plot evolution of widths against analytic comparisons
#        plt.figure(1)
#        plt.plot(self.times, width_z, color="red", ls="-", label="width_z")
#        plt.plot(self.times, width_r, color="blue", linestyle="-", label="width_r")
#        plt.plot(self.times, width_z_ideal, color="red", ls="--", label="width_z_ideal")
#        plt.plot(self.times, width_r_ideal, color="blue", ls="--", label="width_r_ideal")
#        plt.plot(
#            self.times, R_z_TF * TF_z_to_width, color="red", ls=":", label="width_z_TF"
#        )
#        plt.plot(
#            self.times, R_r_TF * TF_r_to_width, color="blue", ls=":", label="width_r_TF"
#        )
#        plt.xlabel("time (oscillator units)")
#        plt.ylabel("widths (oscillator units)")
#        plt.legend()

#        if COM:
#            # plot com in the z-direction
#            plt.figure(2)
#            plt.plot(self.times, com_z, label="com - z")
#            plt.xlabel("time (oscillator units)")
#            plt.ylabel("widths (oscillator units)")
#            plt.legend()

# Animations will come after we get the basics working...
#
#    def animate_density(self, frame_interval=50, rlim=None, zlim=None):
#        """
#        Animates the change in density and potential over time
#        int frame_interval: number of frames to skip each interval, determines smoothness
#        xlim and ylim allow you to zoom in the X or Y directions
#        Returns an animation
#        """
#        fig, ax = plt.subplots()
#        density = self.wavefunction.density()

#        extent = [
#            self.wavefunction.z[0, 0],
#            self.wavefunction.z[0, -1],
#            -self.wavefunction.r[-1, 0],
#            self.wavefunction.r[-1, 0],
#        ]
#        if zlim is not None:
#            extent[0], extent[1] = zlim
#        if rlim is not None:
#            extent[2], extent[3] = rlim

#        # vmin and vmax normalize the colorbar, making the cloud more visible
#        im = ax.imshow(
#            np.concatenate((np.flip(density, axis=0), density)),
#            extent=extent,
#            cmap="nipy_spectral",
#        )  # normalize in here with vmin and vmax
#        cont1 = ax.contour(
#            self.wavefunction.z,
#            self.wavefunction.r,
#            self.potential.harmonicPotential(),
#            colors="white",
#        )
#        cont2 = ax.contour(
#            self.wavefunction.z,
#            -self.wavefunction.r,
#            self.potential.harmonicPotential(),
#            colors="white",
#        )
#        levels = [self.potential.barrier_height / 2]
#        cont3 = ax.contour(
#            self.wavefunction.z,
#            self.wavefunction.r,
#            self.potential.painted_barrier(),
#            levels,
#            colors="red",
#        )
#        cont4 = ax.contour(
#            self.wavefunction.z,
#            -self.wavefunction.r,
#            self.potential.painted_barrier(),
#            levels,
#            colors="red",
#        )
#        plt.colorbar(im, ax=ax, label="density")
#        # cb.remove()
#        # plt.tick_params(left = False, right = False , labelleft = False ,
#        #        labelbottom = False, bottom = False)

#        if zlim is not None:
#            ax.set_ylim(zlim)
#        if rlim is not None:
#            ax.set_xlim(rlim)

#        def update(frame):
#            nonlocal cont1, cont2, cont3, cont4
#            self.potential.adjust_potential(self.times[frame])
#            self.wavefunction.psi = self.psi_history[frame]
#            density = self.wavefunction.density()
#            im.set_array(np.concatenate((np.flip(density, axis=0), density)))
#            for c in cont1.collections:
#                c.remove()
#            for c in cont2.collections:
#                c.remove()
#            for c in cont3.collections:
#                c.remove()
#            for c in cont4.collections:
#                c.remove()
#            cont1 = ax.contour(
#                self.wavefunction.z,
#                self.wavefunction.r,
#                self.potential.harmonicPotential(),
#                colors="white",
#            )
#            cont2 = ax.contour(
#                self.wavefunction.z,
#                -self.wavefunction.r,
#                self.potential.harmonicPotential(),
#                colors="white",
#            )
#            levels = [self.potential.barrier_height / 2]
#            cont3 = ax.contour(
#                self.wavefunction.z,
#                self.wavefunction.r,
#                self.potential.painted_barrier(),
#                levels,
#                colors="red",
#            )
#            cont4 = ax.contour(
#                self.wavefunction.z,
#                -self.wavefunction.r,
#                self.potential.painted_barrier(),
#                levels,
#                colors="red",
#            )
#            #
#            return im, cont1, cont2, cont3, cont4

#        anim = FuncAnimation(
#            fig,
#            update,
#            frames=range(0, len(self.psi_history), frame_interval),
#            interval=50,
#            blit=False,
#        )

#        return anim

#    def animate_profiles(self, frame_interval=50):
#        """
#        Animates the density profiles and change in potentials over time
#        int frame_interval: number of frames to skip each interval, determines smoothness
#        xlim and ylim allow you to zoom in the X or Y directions
#        Returns an animation
#        """
#        fig, (ax1, ax2) = plt.subplots(2, 1)

#        (line1,) = ax1.plot([], [], lw=2, color="b", label="Density Profile - z")
#        (line2,) = ax2.plot([], [], lw=2, color="b", label="Density Profile - r")
#        vl1 = ax1.axvline(color="r")
#        vl2 = ax1.axvline(color="r")
#        line = [line1, line2]

#        ax1.set_ylim(0, 0.5)
#        ax1.set_xlim(-self.wavefunction.Lz / 2, self.wavefunction.Lz / 2)
#        ax2.set_ylim(0, 10)
#        ax2.set_xlim(0, self.wavefunction.Lr)
#        ax1.grid()
#        ax2.grid()
#        ax1.legend()
#        ax1.set_xlabel("z")
#        ax1.set_ylabel("Density Profile - z")
#        ax2.legend()
#        ax2.set_xlabel("r")
#        ax2.set_ylabel("Density Profile - r")

#        def init():
#            line[0].set_data([], [])
#            line[1].set_data([], [])
#            return line

#        def update(frame):
#            self.potential.adjust_potential(self.times[frame])
#            barrier_right = (
#                self.potential.barrier_position + 2.355 * self.potential.barrier_width
#            )  # FWHM
#            barrier_left = (
#                self.potential.barrier_position - 2.355 * self.potential.barrier_width
#            )
#            self.wavefunction.psi = self.psi_history[frame]
#            x1 = self.wavefunction.z[0]
#            x2 = self.wavefunction.r[:, 0]
#            y1, y2 = self.wavefunction.density_profiles
#            line[0].set_data(x1, y1)
#            line[1].set_data(x2, y2)
#            vl1.set_xdata([barrier_right])
#            vl2.set_xdata([barrier_left])
#            return (line,)

#        anim = FuncAnimation(
#            fig,
#            update,
#            init_func=init,
#            frames=range(0, len(self.psi_history), frame_interval),
#            interval=50,
#            blit=False,
#        )

#        return anim



# Copyright 2023 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from oqtant.simulator.three_dim import ThreeDim


class QMPotential:
    """
    'QMPotential' translate Oqtant 'quantum_matter' object into simulation units and uses that objects get_potential function to construct a 2D array used for the potential energy in the simulation
    """

    # natural units of Oqtant
    w = 2 * np.pi * 50  # weak axis trap frequency
    m = 87 * 1.66054e-27  # rb87 mass
    h = 6.626e-34  # planck's constant
    msec = 1e-3
    microns = 1e-6
    khz = 1e3
    hbar = h / 2 / np.pi  # hbar
    # natural units of the simulation (in SI) #CHECKED
    sim_time_unit = 1 / w  # 3.18 milliseconds = 1 simulation time
    sim_length_unit = np.sqrt(hbar / m / w)  # 1.52 microns = 1 simulation length
    sim_energy_unit = hbar * w
    # convert simulation units to Oqtant units # Checked
    sim_length_to_oqt = (
        sim_length_unit / microns
    )  # convert a simulation length to microns.  1.524 microns
    sim_time_to_oqt = (
        sim_time_unit / msec
    )  # convert a simulation time to microseconds.  3.183 mseconds
    oqt_time_to_sim = 1 / sim_time_to_oqt
    w0x = 1.0  # 50hz
    w0r = 8.0  # 400hz

    def __init__(self, quantum_matter):
        self.quantum_matter = quantum_matter
        self.tdm = ThreeDim()
        self.V = None

    # convert units of quantum matter object to simulation relevant quantities
    @property
    def lifetime(self):
        return (
            self.quantum_matter.lifetime * QMPotential.msec / QMPotential.sim_time_unit
        )

    @property
    def time_of_flight(self):
        return (
            self.quantum_matter.time_of_flight
            * QMPotential.msec
            / QMPotential.sim_time_unit
        )

    def update_potential(self, time):
        """
        times within barrier/landscape jobs are automatically converted to simulation units in update_potential
        updates self.V
        potential_1D_x needs only be called every 100 microseconds--potential speedup step.

        Args:
            time:  time (in simulation units)

        Returns:
            None
        """
        # time: sim -> msec # Checked
        time *= QMPotential.sim_time_unit / QMPotential.msec
        # potential in x-direction (in kHz)
        potential_1D_x = np.array(
            self.quantum_matter.get_potential(
                time,
                self.tdm.x_1d * QMPotential.sim_length_unit / QMPotential.microns,
            )
        )
        # potential in z-direction (in sim. units) # Checked
        potential_1D_x *= QMPotential.h * QMPotential.khz / QMPotential.sim_energy_unit
        # potential in r-direction (in sim. units) # Checked
        potential_1D_r = 0.5 * QMPotential.w0r**2 * self.tdm.r_1d**2
        # meshgrid in r and x directions (simulation units)
        pot_x, pot_r = np.meshgrid(potential_1D_x, potential_1D_r)
        # total potential: simulation units
        self.V = pot_x + pot_r



# Copyright 2023 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from scipy.interpolate import RectBivariateSpline, interp1d

from oqtant.simulator.three_dim import ThreeDim


class WaveFunction:
    """
    'WaveFunction' Defines representation for a wavefunction
    """

    def __init__(self):
        self.tdm = ThreeDim()
        initial_psi = self.initial_psi().astype(complex)
        self.psi = self.normalize(initial_psi)

    def initial_psi(self, sigma_x=1, sigma_r=1) -> np.ndarray:
        """
        Defines the initial wave function of the system with
        controllable widths.

        Args:
            sigma_x int:  width in the x direction
            sigma_r int:  width in the r direction

        Returns:
            psi np.ndarray:  initial wave function

        """
        init_x = np.exp(-((self.tdm.x) ** 2) / (2 * sigma_x**2))
        init_r = np.exp(-((self.tdm.r) ** 2) / (2 * sigma_r**2))

        return init_x * init_r

    def normalize(self, psi):
        """
        Normalizes the wave function to the number of atoms
        Applies normalization directly to self.psi

        Args:
            psi np.ndarray:  wavefunction

        Returns:
            None
        """
        B = (
            2
            * np.pi
            * self.tdm.dx
            * self.tdm.dr
            * np.sum(self.tdm.r * np.abs(psi) ** 2)
        )
        # In these units, psi is normalized to unity.  U0 is then the TF parameter N*a/aho.
        return psi * np.sqrt(1 / B)

    @property
    def density(self):
        """
        Returns the density of the wave function
        """
        return np.abs(self.psi) ** 2

    @property
    def column_densities(self):
        """
        Returns the column densities (#/Length^2)
        column_zy:  the zy axes
        column_zx:  the zx axes
        profiles:  the profiles in the x and z axes
        """
        # column_xz
        column_zy = np.zeros((2 * self.tdm.Nr, 2 * self.tdm.Nr))
        profile_r = self.tdm.dx * np.sum(np.abs(self.psi) ** 2, axis=1)  #
        # 1D interpolation
        profile_r_interp = interp1d(
            self.tdm.r[:, 0],
            profile_r,
            fill_value=0.0,
            bounds_error=False,
            kind="cubic",
        )

        for iz in range(len(self.tdm.z[0, :])):
            for iy in range(len(self.tdm.y[:, 0])):
                r = np.sqrt(self.tdm.z[0, iz] ** 2 + self.tdm.y[iy, 0] ** 2)
                column_zy[iz, iy] = profile_r_interp(r)

        # column zy
        # 2D interpolation
        density_interp = RectBivariateSpline(
            self.tdm.r[:, 0], self.tdm.x[0, :], self.density, kx=3, ky=3
        )  # cubic spline interpolation
        column_zx = np.zeros((2 * self.tdm.Nr, self.tdm.Nx))

        for iz in range(len(self.tdm.z[0, :])):
            for ix in range(len(self.tdm.x[0, :])):
                r = np.sqrt(self.tdm.z[0, iz] ** 2 + self.tdm.y[:, 0] ** 2)
                density_slice = density_interp(np.sort(r), self.tdm.x[0, ix])
                column_zx[iz, ix] = self.tdm.dy * np.sum(density_slice)

        #
        profile_z_zy = self.tdm.dy * np.sum(column_zy, axis=1)
        profile_z_zx = self.tdm.dx * np.sum(column_zx, axis=1)
        profile_x = self.tdm.dz * np.sum(column_zx, axis=0)
        profiles = [profile_z_zy, profile_z_zx, profile_x]
        # check that interpolation preserved norm
        if (
            self.tdm.dy * sum(profile_z_zy) > 0.95
            and self.tdm.dx * sum(profile_x) > 0.95
        ):
            pass
        else:
            raise ValueError("Lost norm in interpolation")

        # the profile_x should match up with the output of the profile function.
        return (
            column_zy,
            column_zx,
            profiles,
        )

    @property
    def density_profiles(self):
        """
        Returns the density profiles along the x and r axes.  #/Length
        These match the integrated column densities
        """
        profile_x = (
            2 * np.pi * self.tdm.dr * np.sum(self.tdm.r * np.abs(self.psi) ** 2, axis=0)
        )
        profile_r = (
            2
            * np.pi
            * self.tdm.dx
            * self.tdm.r[:, 0]
            * np.sum(np.abs(self.psi) ** 2, axis=1)
        )
        return profile_x, profile_r

    @property
    def atom_number(self):
        """
        Returns atom number
        """
        return (
            2
            * np.pi
            * self.tdm.dx
            * self.tdm.dr
            * np.sum(self.tdm.r * np.abs(self.psi) ** 2)
        )

    @property
    def phase(self):
        """
        Returns the phase of the wave function
        """
        return np.angle(self.psi)

    @property
    def flow(self):
        """
        Returns the superfluid velocities in X and R directions
        Returns two 2D arrays
        """
        diff_r = (
            2 * self.tdm.dr * np.gradient(self.phase, self.tdm.dr, axis=0, edge_order=1)
        )  # axis 0 for r
        diff_x = (
            2 * self.tdm.dz * np.gradient(self.phase, self.tdm.dz, axis=1, edge_order=1)
        )  # axis 1 for x

        # dealing with 'jumps' from negative pi to pi.  Essentially enforcing that diff_i resides between -pi and pi
        diff_r[diff_r > np.pi] -= 2 * np.pi
        diff_r[diff_r < -np.pi] += 2 * np.pi
        diff_x[diff_x > np.pi] -= 2 * np.pi
        diff_x[diff_x < -np.pi] += 2 * np.pi

        flow_r = 2 * np.pi * diff_r / (2 * self.tdm.dr)
        flow_x = 2 * np.pi * diff_x / (2 * self.tdm.dx)

        return flow_r, flow_x

    @property
    def current(self):
        """
        Returns the total current along the X-direction
        Returns a 1D array
        """
        _, flow_x = self.flow  # v_z(r,x)
        current_x = (
            2 * np.pi * self.tdm.dr * np.sum(self.tdm.r * flow_x * self.density, axis=0)
        )  # j_z(z):  A 1D array of current along the x-direction.
        return current_x

    @property
    def com_position(self):
        """
        Returns the center of mass coordinates of the cloud in X direction
        The center of mass cannot be displaced in the radial direction by assumption
        Returns 1 scalar
        Useful for diagnostics
        """
        com_x = (
            2
            * np.pi
            * self.tdm.dx
            * self.tdm.dr
            * np.sum(self.tdm.x * self.tdm.r * np.abs(self.psi) ** 2)
        )

        return com_x

    @property
    def widths(self):
        """
        Returns the widths (\\Delta r and \\Delta x) in radial and x directions
        Returns 2 scalars
        """
        square_width_x = (
            2
            * np.pi
            * self.tdm.dr
            * self.tdm.dx
            * np.sum(self.tdm.r * self.tdm.x**2 * np.abs(self.psi) ** 2)
        )
        square_width_r = (
            2
            * np.pi
            * self.tdm.dr
            * self.tdm.dx
            * np.sum(self.tdm.r**3 * np.abs(self.psi) ** 2)
        )
        com_x = self.com_position()

        return np.sqrt(square_width_x - com_x**2), np.sqrt(square_width_r)



# Copyright 2023 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
from collections.abc import Callable

import matplotlib.pyplot as plt
import numpy as np
from bert_schemas import job as job_schema

from oqtant.schemas.interpolation import interpolate_1d, interpolate_1d_list

# Module containing methods and classes for abstractions of quasi-1D "painted" light Oqtant object

# "ideal" optical potentials are those specified by the user using included objects
# "actual" optical potentials include implementation and hardware realities
# such as objects being projected as a sum of gaussians on a pre-defined position grid
# with dynamic weights recalculated on a periodic basis


def gaussian(
    xs: np.ndarray,
    amp: float = 1.0,
    center: float = 0.0,
    sigma: float = 1.0,
    offset: float = 0.0,
) -> np.ndarray:
    """Method that evaluates a standard gaussian form over the given input points

    Args:
        xs (numpy.ndarray): Positions where the gaussian should be evaluated
        amp (float, optional): Gaussian amplitude
        center (float, optional): Gaussian center
        sigma (float, optional): Gaussian width
        offset (float, optional): Gaussian dc offset

    Returns:
        np.ndarray: Gaussian function evaluated over the input points
    """
    return amp * np.exp(-((xs - center) ** 2) / (2 * sigma**2)) + offset


class Projected:
    """A class that captures the features, and limitations, of optical objects
    implemented by the Oqtant hardware projection system.
    """

    RESOLUTION = 2.2  # 1/e^2 diameter of projection system, microns
    POSITION_STEP = 1.0  # grid step between projected spots, microns
    POSITION_MIN = -60.0  # minimum position of projected light, microns
    POSITION_MAX = 60  # maximum position of projected light, microns
    PROJECTED_SPOTS = np.arange(POSITION_MIN, POSITION_MAX + 1.0, POSITION_STEP)
    UPDATE_PERIOD = 0.1  # milliseconds between updates of projected light
    ENERGY_MIN = 0.0  # minimum projected energy shift at any position, kHz
    ENERGY_MAX = 100  # maximum projected energy shift at any position, kHz

    @staticmethod
    def get_corrected_times(times: list[float]) -> list[float]:
        """Method to calculate the effective times realized by the projection system,
        which only updates optical features periodically

        Args:
            times (list[float]): Time, in ms, to be corrected

        Returns:
            list[float]: The corrected times
        """
        times_corrected = (
            np.floor((1000.0 * np.asarray(times)) / (1000.0 * Projected.UPDATE_PERIOD))
            * Projected.UPDATE_PERIOD
        )
        return list(times_corrected)

    @staticmethod
    def get_corrected_time(time: float) -> float:
        """Method to calculate the effective time realized by the projection system,
        which only updates optical features periodically

        Args:
            time (float): Time, in ms, to be corrected

        Returns:
            float: The corrected time
        """
        return Projected.get_corrected_times(times=[time])[0]

    # gets corrected weights at each projected gaussian spot to reproduce the ideal
    # potential energy vs position as closely as is reasonable
    @staticmethod
    def get_projection_weights(
        get_ideal_potential: Callable[[float], list], time: float = 0
    ) -> list[float]:
        """Method to calculate weights for each horizontal "spot" projected onto the atom ensemble to
        attempt to achieve the passed optical object's "ideal" potential energy profile.
        Implements first-order corrections for anamolous contributions from nearby spots,
        inter-integer barrier centers, etc

        Args:
            get_ideal_potential (Callable[[float], list]): Method for the optical object or any class
                that supports optical objects that calculates the specified "ideal" or "requested"
                potential energy profile
            time (float, optional): Time at which to correct

        Returns:
            list[float]: Calculated (optical intensity) contribution for each projected spot
                (diffraction frequency) used by the projection systems
        """

        positions_fine = np.arange(
            Projected.POSITION_MIN, Projected.POSITION_MAX + 0.1, 0.1
        )

        # calculate the ideal potential over the entire spatial region
        potential_ideal = np.asarray(
            get_ideal_potential(time=time, positions=positions_fine)
        )

        # calculate the optical field that would result from raw object data
        weights = np.asarray(
            get_ideal_potential(time=time, positions=Projected.PROJECTED_SPOTS)
        )
        potential_actual = np.zeros_like(positions_fine)
        for indx, spot in enumerate(Projected.PROJECTED_SPOTS):
            potential_actual += gaussian(
                xs=positions_fine,
                amp=weights[indx],
                center=spot,
                sigma=Projected.RESOLUTION / 4.0,
                offset=0.0,
            )

        # recompute weights with overall scaling to achieve correct peak height/energy
        # this removes first order variation in height with inter-grid object centers
        # and contributions from adjacent space/frequency spots
        maximum = max(potential_actual)
        scaling = max(potential_ideal) / maximum if maximum > 0.0 else 0.0
        return list(scaling * weights)

    @staticmethod
    def get_actual_potential(
        get_ideal_potential: Callable[[float], list],
        time: float = 0.0,
        positions: list = PROJECTED_SPOTS,
    ) -> list[float]:
        """Method to calculate the "actual" potential energy vs position for optical
        objects/fields as realized by the Oqtant projection system. Includes effects,
        and first-order corrections for, finite time updates and finite optical
        resolution/optical objects being projected as sums of gaussians and energetic
        clipping of optical potentials at 100 kHz

        Args:
            get_ideal_potential (Callable[[float], list]): Object method for request/ideal potential
            time (float, optional): Time to evaluate ideal potential
            positions (list[float], optional): Positions to evaluate the actual potential at

        Returns:
            list[float]: Expected actual potential energy at the request positions
        """
        time = Projected.get_corrected_time(time)  # include finite-update period
        weights = Projected.get_projection_weights(get_ideal_potential, time)
        potential = np.zeros_like(positions)
        for indx, spot in enumerate(Projected.PROJECTED_SPOTS):
            potential += gaussian(
                xs=positions,
                amp=weights[indx],
                center=spot,
                sigma=Projected.RESOLUTION / 4.0,
                offset=0.0,
            )
        return list(np.clip(potential, Projected.ENERGY_MIN, Projected.ENERGY_MAX))


class Snapshot(job_schema.Landscape):
    """A class that represents a painted optical landscape/potential at a single
    point in (manipulation stage) time
    """

    @classmethod
    def new(
        cls,
        time: float = 0,
        positions: list = [-10, 10],
        potentials: list = [0, 0],
        interpolation: job_schema.InterpolationType = "LINEAR",
    ) -> Snapshot:
        """Method to create a new Snapshot object

        Args:
            time (float, optional): Time associated with the snapshot
            positions (list, optional): Position list for the snapshot
            potentials (list, optional): Potential energies corresponding to the list of positions
            interpolation (bert_schemas.job.InterpolationType, optional): How to connect the object's
                (positions, potentials) data in space.

        Returns:
            Snapshot: a new Snapshot object
        """
        return cls(
            time_ms=time,
            positions_um=positions,
            potentials_khz=potentials,
            spatial_interpolation=interpolation,
        )

    @classmethod
    def from_input(cls, landscape: job_schema.Landscape) -> Snapshot:
        """Method to create a Snapshot object from an existing jobs input

        Args:
            landscape (bert_schemas.job.Landscape): The input values

        Returns:
            Snapshot: A new Snapshot object created using the input data
        """
        return cls(**landscape.model_dump())

    def get_ideal_potential(
        self, time=0.0, positions: list = Projected.PROJECTED_SPOTS
    ) -> list[float]:
        """Method to get the ideal potential energy at the specified positions

        Args:
            positions (list, optional): List of positions in microns

        Returns:
            list[float]: Potential energies, in kHz, at the specified positions
        """
        potentials = interpolate_1d_list(
            self.positions_um,
            self.potentials_khz,
            positions,
            self.spatial_interpolation,
        )
        return potentials

    def get_potential(self, positions: list[float]) -> list[float]:
        """Method to calculate the optical potential associated with a Landscape object,
        taking into account the actual implementation of the Oqtant projection system,
        at the given time

        Args:
            positions (list[float]): Positions, in microns, where the potential should be evaluated

        Returns:
            list[float]: Potential energies, in kHz, at the specified positions
        """
        return Projected.get_actual_potential(
            self.get_ideal_potential, positions=positions
        )

    def show_potential(
        self,
        xlimits: list[float] = [Projected.POSITION_MIN - 1, Projected.POSITION_MAX + 1],
        ylimits: list[float] = [Projected.ENERGY_MIN - 1, Projected.ENERGY_MAX + 1],
        include_ideal: bool = False,
    ) -> None:
        """Method to plot the potential energy as a function of position for a Landscape
        object at the given times

        Args:
            xlimits (list[float], optional): Plot limits for x axis
            ylimits (list[float], optional): Plot limits for y axis
            include_ideal (bool, optional): Flag for including target potential in plot
        """

        positions = np.arange(
            Projected.POSITION_MIN, Projected.POSITION_MAX + 0.1, 0.1, dtype=float
        )

        _, ax = plt.subplots()
        color = next(ax._get_lines.prop_cycler)["color"]
        (ln,) = plt.plot(
            positions, self.get_potential(positions=positions), color=color
        )
        if include_ideal:
            (ln2,) = plt.plot(
                positions,
                self.get_ideal_potential(positions=positions),
                "--",
                color=color,
            )
            plt.plot(self.positions_um, self.potentials_khz, ".", color=color)
            lns = [ln, ln2]
            labs = ["actual", "ideal"]
            ax.legend(lns, labs, loc=0)
        plt.xlabel("position (microns)", labelpad=6)
        plt.ylabel("potential energy (kHz)", labelpad=6)
        plt.title("Snapshot potential energy profile")
        plt.xlim(xlimits)
        plt.ylim(ylimits)
        plt.show()


# (potentially) dynamic landscape made up of snapshots
class Landscape(job_schema.OpticalLandscape):
    """Class that represents a dynamic painted-potential optical landscape constructed
    from individual (instantaneous time) Snapshots
    """

    @classmethod
    def new(
        cls,
        snapshots: list[Snapshot] = [
            Snapshot.new(time=0),
            Snapshot.new(time=2),
        ],
    ) -> Landscape:
        """Method to create a new Landscape object

        Args:
            snapshots (list[Snapshot], optional): A list of Snapshot objects

        Returns:
            Landscape: A new Landscape object
        """
        optical_landscapes = []
        for snapshot in snapshots:
            optical_landscapes.append(
                job_schema.Landscape(
                    time_ms=snapshot.time_ms,
                    positions_um=snapshot.positions_um,
                    potentials_khz=snapshot.potentials_khz,
                    spatial_interpolation=snapshot.spatial_interpolation,
                )
            )
        return cls(landscapes=optical_landscapes)

    @classmethod
    def from_input(cls, landscape: job_schema.OpticalLandscape) -> Landscape:
        """Method to create a Landscape object from an existing jobs input

        Args:
            landscape (job_schema.OpticalLandscape): The input values

        Returns:
            Landscape: A new Landscape object
        """
        return cls(**json.loads(landscape.model_dump_json()))

    # extract Snapshot abstract objects from backend data structure
    @property
    def snapshots(self) -> list[Snapshot]:
        """Property to get a list of Snapshot objects associated to a Landscape object

        Returns:
            list[Snapshot]: List of Snapshot objects
        """
        return [Snapshot(**landscape.model_dump()) for landscape in self.landscapes]

    def get_ideal_potential(
        self, time: float, positions: list[float] = Projected.PROJECTED_SPOTS
    ) -> list[float]:
        """Method to calculate ideal object potential energy at the specified time and positions

        Args:
            time (float): Time, in ms, at which the potential energy is calculated
            positions (list[float], optional): Positions at which the potential energy is calculated

        Returns:
            list[float]: Potential energies, in kHz, at specified time and positions
        """
        potential = [0] * len(positions)
        snaps = self.snapshots
        if len(snaps) < 2:
            return potential
        snap_times = [snap.time_ms for snap in snaps]
        if time >= min(snap_times) and time <= max(snap_times):
            pre = next(snap for snap in reversed(snaps) if snap.time_ms <= time)
            nex = next(snap for snap in snaps if snap.time_ms >= time)
            potential = [
                interpolate_1d(
                    [pre.time_ms, nex.time_ms], [p1, p2], time, self.interpolation
                )
                for p1, p2 in zip(
                    pre.get_ideal_potential(positions=positions),
                    nex.get_ideal_potential(positions=positions),
                )
            ]
        return potential

    def get_potential(
        self, time: float, positions: list = Projected.PROJECTED_SPOTS
    ) -> list[float]:
        """Method to calculate the optical potential associated with a Landscape object,
        taking into account the actual implementation of the Oqtant projection system,
        at the given time

        Args:
            time (float): Time, in ms, at which to sample the potential energy
            positions (list[float], optional): Positions, in microns, where the potential should be evaluated

        Returns:
            list[float]: Potential energies, in kHz, at the requested positions and time
        """
        return Projected.get_actual_potential(
            self.get_ideal_potential, time=time, positions=positions
        )

    def show_potential(
        self,
        times: list = [0.0],
        xlimits: list = [Projected.POSITION_MIN - 1, Projected.POSITION_MAX + 1],
        ylimits: list = [Projected.ENERGY_MIN - 1, Projected.ENERGY_MAX + 1],
        include_ideal: bool = False,
    ):
        """Method to plot the potential energy as a function of position for a Landscape object at the given times

        Args:
            times (list[float], optional): Times, in ms, at which to evaluate and plot the potential
            xlimits (list[float], optional): Plot limits for x axis
            ylimits (list[float], optional): Plot limits for y axis
            include_ideal (bool, optional): Flag for including target potential in plot
        """
        positions = np.arange(
            Projected.POSITION_MIN, Projected.POSITION_MAX + 0.1, 0.1, dtype=float
        )

        fig, ax = plt.subplots()
        lns = []
        labs = []
        for time in times:
            potentials = self.get_potential(time, positions)
            color = next(ax._get_lines.prop_cycler)["color"]
            (ln,) = plt.plot(positions, potentials, color=color)
            lns.append(ln)
            labs.append("t = " + str(time) + " ms")
            if include_ideal:
                potentials_ideal = self.get_ideal_potential(
                    time=time, positions=positions
                )
                (ln,) = plt.plot(positions, potentials_ideal, "--", color=color)
                lns.append(ln)
                labs.append("t = " + str(time) + " ms (ideal)")

        plt.xlabel("position (microns)", labelpad=6)
        plt.ylabel("potential energy (kHz)", labelpad=6)
        plt.title("Landscape potential energy profile")
        plt.xlim(xlimits)
        plt.ylim(ylimits)
        ax.legend(lns, labs, loc=0)
        plt.show()


class Barrier(job_schema.Barrier):
    """Class that represents a painted optical barrier."""

    @classmethod
    def new(
        cls,
        positions: list[float] = [0.0, 0.0],
        heights: list[float] = [0.0, 0.0],
        widths: list[float] = [1.0, 1.0],
        times: list[float] = [0.0, 10.0],
        shape: job_schema.ShapeType = job_schema.ShapeType.GAUSSIAN,
        interpolation: job_schema.InterpolationType = job_schema.InterpolationType.LINEAR,
    ) -> Barrier:
        """Method to create a new Barrier object

        Args:
            positions (list[float], optional): Positions for the barrier
            heights (list[float], optional): Heights for the barrier
            widths (list[float], optional): Widths for the barrier
            times (list[float], optional): Times for the barrier
            shape (bert_schemas.job.ShapeType, optional): Shape of the barrier
            interpolation (bert_schemas.job.InterpolationType, optional): Interpolation type of the barrier

        Returns:
            Barrier: A new Barrier object

        Raises:
            ValueError: if data lists are not of equal length
        """
        if not (len(positions) == len(heights) == len(widths) == len(times)):
            raise ValueError(
                "Barrier data lists must be of equal length, default minimum of 2"
            )

        data = {
            "times_ms": times,
            "positions_um": positions,
            "heights_khz": heights,
            "widths_um": widths,
            "shape": shape,
            "interpolation": interpolation,
        }

        return cls(**data)

    @classmethod
    def from_input(cls, barrier: job_schema.Barrier) -> Barrier:
        """Method to create a Barrier object using the input values of a job

        Args:
            barrier (job_schema.Barrier): The input values

        Returns:
            Barrier: A new Barrier object created using the input data
        """
        return cls(**barrier.model_dump())

    @property
    def birth(self) -> float:
        """Property to get the (manipulation stage) time that the Barrier object will be created

        Returns:
            float: The time, in ms, at which the barrier will start being projected
        """
        return min(self.times_ms)

    @property
    def death(self) -> float:
        """Property to get the (manipulation stage) time that the Barrier object will cease to exist

        Returns:
            float: The time, in ms, at which the barrier will stop being projected
        """
        return max(self.times_ms)

    def evolve(
        self,
        duration: float,
        position: float = None,
        height: float = None,
        width: float = None,
    ) -> None:
        """Method to evolve the position, height, and/or width of a Barrier object over a duration

        Args:
            duration (float): The time, in ms, over which evolution should take place
            position (float | None, optional): The position, in microns, to evolve to
            height (float | None, optional): The height, in kHz, to evolve to
            width (float | None, optional): The width, in microns, to evolve to
        """
        if position is None:
            position = self.positions_um[-1]
        if height is None:
            height = self.heights_khz[-1]
        if width is None:
            width = self.widths_um[-1]
        self.positions_um.append(position)
        self.heights_khz.append(height)
        self.widths_um.append(width)
        self.times_ms.append(self.times_ms[-1] + duration)

    def is_active(self, time: float) -> bool:
        """Method to determine if a Barrier object is active (exists) at the specified time

        Args:
            time (float): The time, in ms, at which the query is evaluated

        Returns:
            bool: Flag indicating if the barrier exists or not at the specified time
        """
        return time >= self.times_ms[0] and time <= self.times_ms[-1]

    def get_positions(self, times: list[float]) -> list[float]:
        """Method to calculate the Barrier object position at the specified (manipulation stage) times

        Args:
            times (list[float]): The times, in ms, at which positions are calculated

        Returns:
            list[float]: The positions, in microns, at the specified times
        """
        return interpolate_1d_list(
            self.times_ms,
            self.positions_um,
            Projected.get_corrected_times(times=times),
            self.interpolation,
        )

    def get_position(self, time: float) -> float:
        """Method to calculate the Barrier object position at the specified (manipulation stage) time

        Args:
            time (float): The time, in ms, at which the position is calculated

        Returns:
            float: The position, in microns, at the specified time
        """
        return self.get_positions(times=[time])[0]

    def get_heights(self, times: list[float]) -> list[float]:
        """Method to calculate the Barrier object heights at the specified list of times

        Args:
            times (list[float]): The times, in ms, at which the heights are calculated

        Returns:
            list[float]: The barrier heights at the specified times
        """
        return interpolate_1d_list(
            self.times_ms,
            self.heights_khz,
            Projected.get_corrected_times(times=times),
            self.interpolation,
        )

    def get_height(self, time: float) -> float:
        """Method to get the Barrier object height at the specified time

        Args:
            time (float): The time, in ms, at which the height is calculated

        Returns:
            float: The barrier height at the specified time
        """
        return self.get_heights(times=[time])[0]

    def get_widths(self, times: list[float]) -> list[float]:
        """Method to calculate the Barrier object widths at the specified times

        Args:
            times (list[float]): The times, in ms, at which the heights are calculated

        Returns:
            list[float]: The barrier widths at the specified times
        """
        return interpolate_1d_list(
            self.times_ms,
            self.widths_um,
            Projected.get_corrected_times(times=times),
            self.interpolation,
        )

    def get_width(self, time: float) -> float:
        """Method to calculate the Barrier object width at the specified time

        Args:
            times (float): The time, in ms, at which the height is calculated

        Returns:
            float: The barrier width at the specified time
        """
        return self.get_widths(times=[time])[0]

    def get_ideal_potential(
        self, time: float = 0.0, positions: list[float] = Projected.PROJECTED_SPOTS
    ) -> list[float]:
        """Method to calculate the ideal Barrier object potential energy at the given positions
        and at the specified time without taking into account finite projection system resolution
        to update time of projected light

        Args:
            time (float, optional): The time, in ms, at which the potential is calculated
            positions (list[float], optional): The positions, in microns, at which the potential
                energies are evaluated

        Returns:
            list[float]: The potential energies, in kHz, at the specified positions
        """
        h = self.get_height(time)
        p = self.get_position(time)
        w = self.get_width(time)
        potential = [0] * len(positions)
        if h <= 0 or w <= 0 or not self.is_active(time):
            return potential
        if self.shape == "SQUARE":  # width = half width
            potential = [0 if (x < p - w or x > p + w) else h for x in positions]
        elif self.shape == "LORENTZIAN":  # width == HWHM (half-width half-max)
            potential = [h / (1 + ((x - p) / w) ** 2) for x in positions]
        elif self.shape == "GAUSSIAN":  # width = sigma (Gaussian width)
            potential = [h * np.exp(-((x - p) ** 2) / (2 * w**2)) for x in positions]
        return potential

    def get_potential(
        self, time: float, positions: list[float] = Projected.PROJECTED_SPOTS
    ) -> list[float]:
        """Method to calculate the optical potential associated with a Barrier object, taking into
        account the actual implementation of the Oqtant projection system

        Args:
            time (float): The time, in ms, at which the potential should be evaluated
            positions (list[float], optional): The positions, in microns, at which the potential should be evaluated

        Returns:
            list[float]: The potential energies, in kHz, at the specified positions
        """
        return Projected.get_actual_potential(
            self.get_ideal_potential, time=time, positions=positions
        )

    def show_dynamics(self) -> None:
        """Method to plot the position, width and height of a Barrier object over time"""
        tstart = min(self.times_ms)
        tstop = max(self.times_ms)
        times = np.linspace(
            tstart,
            tstop,
            num=int((tstop - tstart) / Projected.UPDATE_PERIOD),
            endpoint=True,
        )
        fig, ax1 = plt.subplots()

        # plot position and width vs time
        style = "steps-pre"
        color = next(ax1._get_lines.prop_cycler)["color"]
        ax1.set_xlabel("time (ms)")
        ax1.set_ylabel("position or width (microns)")
        ax1.set_xlim([-1, self.times_ms[-1] + 1])
        ax1.set_ylim([Projected.POSITION_MIN - 1, Projected.POSITION_MAX + 1])
        (ln1,) = plt.plot(
            times, self.get_positions(times), color=color, drawstyle=style
        )
        plt.plot(
            self.times_ms,
            self.get_positions(self.times_ms),
            ".",
            color=color,
        )
        color = next(ax1._get_lines.prop_cycler)["color"]
        (ln2,) = plt.plot(times, self.get_widths(times), color=color, drawstyle=style)
        plt.plot(self.times_ms, self.get_widths(self.times_ms), ".", color=color)

        # plot height on the same time axis
        ax2 = ax1.twinx()
        ax2.set_ylabel("height (kHz)")
        ax2.set_ylim([0, 100])
        color = next(ax1._get_lines.prop_cycler)["color"]
        (ln3,) = plt.plot(times, self.get_heights(times), color=color, drawstyle=style)
        plt.plot(self.times_ms, self.get_heights(self.times_ms), ".", color=color)

        # shared setup
        color = next(ax1._get_lines.prop_cycler)["color"]
        ax1.legend([ln1, ln2, ln3], ["position", "width", "height"], loc="upper left")
        plt.title("Barrier dynamics")
        fig.tight_layout()
        plt.show()

    def show_potential(
        self,
        times: list[float] = [0.0],
        xlimits: list[float] = [Projected.POSITION_MIN - 1, Projected.POSITION_MAX + 1],
        ylimits: list[float] = [Projected.ENERGY_MIN - 1, Projected.ENERGY_MAX + 1],
        include_ideal: bool = False,
    ) -> None:
        """Method to plot the potential energy as a function of position for a Barrier object

        Args:
            times (list[float], optional): The times, in ms, at which the potential is evaluated
            xlimits (list[float], optional): Plot limits for x axis
            ylimits (list[float], optional): Plot limits for y axis
            include_ideal (bool, optional): Flag for including target potential in plot
        """

        positions = np.arange(np.floor(min(xlimits)), np.ceil(max(xlimits)) + 1, 0.1)

        fig, ax1 = plt.subplots()
        ax = plt.gca()
        lns = []
        labs = []
        for time in times:
            color = next(ax._get_lines.prop_cycler)["color"]
            potential = self.get_potential(time=time, positions=positions)
            (ln,) = plt.plot(positions, potential, color=color)
            lns.append(ln)
            labs.append("t = " + str(time) + " ms")
            if include_ideal:
                potentials_ideal = self.get_ideal_potential(
                    time=time, positions=positions
                )
                (ln,) = plt.plot(positions, potentials_ideal, "--", color=color)
                lns.append(ln)
                labs.append("t = " + str(time) + "ms (ideal)")

        plt.xlabel("position (microns)", labelpad=6)
        plt.ylabel("potential energy (kHz)", labelpad=6)
        plt.xlim(xlimits)
        plt.ylim(ylimits)
        ax1.legend(lns, labs, loc=0)
        plt.show()



# Copyright 2023 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import threading
import time

import uvicorn


class ThreadServer(uvicorn.Server):
    """Class for spinning up a server in a thread"""

    def install_signal_handlers(self) -> None:
        ...

    @contextlib.contextmanager
    def run_in_thread(self) -> None:
        """Method to create a thread and close once finished"""
        thread = threading.Thread(target=self.run)
        thread.start()
        try:
            while not self.started:
                time.sleep(1e-3)
            yield
        finally:
            self.should_exit = True
            thread.join()



# Copyright 2023 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

import numpy as np


@dataclass
class ThreeDim:
    """
    'ThreeDim' Defines a two dimensional grid space in cylindrical coordinates with axial symmetry.
    Nx and Nr are the number of points in the x and r directions
    Lx and Lr are the lenghts of the x and z dimensions
    """

    # parameters for 3D space
    Nx = 300
    Nr = 100
    Lx = 30
    Lr = 4
    dx = Lx / Nx
    dr = Lr / Nr

    # defining cylindrical grids using np.meshgrid
    x_1d, r_1d = np.linspace(-Lx / 2, Lx / 2, Nx), np.linspace(dr / 2, Lr - dr / 2, Nr)

    x, r = np.meshgrid(x_1d, r_1d)

    # defining the equivalent cartesian grids for constructing the column densities
    z, y = np.meshgrid(
        np.linspace(-Lr + dr / 2, Lr - dr / 2, 2 * Nr),
        np.linspace(-Lr + dr / 2, Lr - dr / 2, 2 * Nr),
    )
    dz, dy = dr, dr
